@@ -62,6 +62,7 @@
 #include "gdaemonfileinputstream.h"
 #include "gvfsdaemondbus.h"
 #include <gvfsdaemonprotocol.h>
+#include <gvfsfileinfo.h>
 
 #define MAX_READ_SIZE (4*1024*1024)
 
@@ -151,6 +152,30 @@ typedef struct {
   guint32 seq_nr;
 } CloseOperation;
 
+typedef enum {
+  QUERY_STATE_INIT = 0,
+  QUERY_STATE_WROTE_REQUEST,
+  QUERY_STATE_HANDLE_INPUT,
+  QUERY_STATE_HANDLE_INPUT_BLOCK,
+  QUERY_STATE_HANDLE_HEADER,
+  QUERY_STATE_READ_BLOCK,
+  QUERY_STATE_SKIP_BLOCK
+} QueryState;
+
+typedef struct {
+  QueryState state;
+
+  /* Input */
+  char *attributes;
+  
+  /* Output */
+  GFileInfo *info;
+  GError *ret_error;
+
+  gboolean sent_cancel;
+  
+  guint32 seq_nr;
+} QueryOperation;
 
 typedef struct {
   gboolean cancelled;
@@ -164,7 +189,15 @@ typedef struct {
   gboolean io_cancelled;
 } IOOperationData;
 
-typedef StateOp (*state_machine_iterator) (GDaemonFileInputStream *file, IOOperationData *io_op, gpointer data);
+typedef struct {
+  char *data;
+  gsize len;
+  int seek_generation;
+} PreRead;
+
+typedef StateOp (*state_machine_iterator) (GDaemonFileInputStream *file,
+					   IOOperationData *io_op,
+					   gpointer data);
 
 struct _GDaemonFileInputStream {
   GFileInputStream parent;
@@ -177,6 +210,8 @@ struct _GDaemonFileInputStream {
   guint32 seq_nr;
   goffset current_offset;
 
+  GList *pre_reads;
+  
   InputState input_state;
   gsize input_block_size;
   int input_block_seek_generation;
@@ -185,61 +220,89 @@ struct _GDaemonFileInputStream {
   GString *output_buffer;
 };
 
-static gssize     g_daemon_file_input_stream_read          (GInputStream         *stream,
-							    void                 *buffer,
-							    gsize                 count,
-							    GCancellable         *cancellable,
-							    GError              **error);
-static gssize     g_daemon_file_input_stream_skip          (GInputStream         *stream,
-							    gsize                 count,
-							    GCancellable         *cancellable,
-							    GError              **error);
-static gboolean   g_daemon_file_input_stream_close         (GInputStream         *stream,
-							    GCancellable         *cancellable,
-							    GError              **error);
-static GFileInfo *g_daemon_file_input_stream_query_info    (GFileInputStream     *stream,
-							    char                 *attributes,
-							    GCancellable         *cancellable,
-							    GError              **error);
-static goffset    g_daemon_file_input_stream_tell          (GFileInputStream     *stream);
-static gboolean   g_daemon_file_input_stream_can_seek      (GFileInputStream     *stream);
-static gboolean   g_daemon_file_input_stream_seek          (GFileInputStream     *stream,
-							    goffset               offset,
-							    GSeekType             type,
-							    GCancellable         *cancellable,
-							    GError              **error);
-static void       g_daemon_file_input_stream_read_async    (GInputStream         *stream,
-							    void                 *buffer,
-							    gsize                 count,
-							    int                   io_priority,
-							    GCancellable         *cancellable,
-							    GAsyncReadyCallback   callback,
-							    gpointer              data);
-static gssize     g_daemon_file_input_stream_read_finish   (GInputStream         *stream,
-							    GAsyncResult         *result,
-							    GError              **error);
-static void       g_daemon_file_input_stream_skip_async    (GInputStream         *stream,
-							    gsize                 count,
-							    int                   io_priority,
-							    GCancellable         *cancellable,
-							    GAsyncReadyCallback   callback,
-							    gpointer              data);
-static gssize     g_daemon_file_input_stream_skip_finish   (GInputStream         *stream,
-							    GAsyncResult         *result,
-							    GError              **error);
-static void       g_daemon_file_input_stream_close_async   (GInputStream         *stream,
-							    int                   io_priority,
-							    GCancellable         *cancellable,
-							    GAsyncReadyCallback   callback,
-							    gpointer              data);
-static gboolean   g_daemon_file_input_stream_close_finish  (GInputStream         *stream,
-							    GAsyncResult         *result,
-							    GError              **error);
+static gssize     g_daemon_file_input_stream_read              (GInputStream         *stream,
+								void                 *buffer,
+								gsize                 count,
+								GCancellable         *cancellable,
+								GError              **error);
+static gssize     g_daemon_file_input_stream_skip              (GInputStream         *stream,
+								gsize                 count,
+								GCancellable         *cancellable,
+								GError              **error);
+static gboolean   g_daemon_file_input_stream_close             (GInputStream         *stream,
+								GCancellable         *cancellable,
+								GError              **error);
+static GFileInfo *g_daemon_file_input_stream_query_info        (GFileInputStream     *stream,
+								const char           *attributes,
+								GCancellable         *cancellable,
+								GError              **error);
+static goffset    g_daemon_file_input_stream_tell              (GFileInputStream     *stream);
+static gboolean   g_daemon_file_input_stream_can_seek          (GFileInputStream     *stream);
+static gboolean   g_daemon_file_input_stream_seek              (GFileInputStream     *stream,
+								goffset               offset,
+								GSeekType             type,
+								GCancellable         *cancellable,
+								GError              **error);
+static void       g_daemon_file_input_stream_read_async        (GInputStream         *stream,
+								void                 *buffer,
+								gsize                 count,
+								int                   io_priority,
+								GCancellable         *cancellable,
+								GAsyncReadyCallback   callback,
+								gpointer              data);
+static gssize     g_daemon_file_input_stream_read_finish       (GInputStream         *stream,
+								GAsyncResult         *result,
+								GError              **error);
+static void       g_daemon_file_input_stream_skip_async        (GInputStream         *stream,
+								gsize                 count,
+								int                   io_priority,
+								GCancellable         *cancellable,
+								GAsyncReadyCallback   callback,
+								gpointer              data);
+static gssize     g_daemon_file_input_stream_skip_finish       (GInputStream         *stream,
+								GAsyncResult         *result,
+								GError              **error);
+static void       g_daemon_file_input_stream_close_async       (GInputStream         *stream,
+								int                   io_priority,
+								GCancellable         *cancellable,
+								GAsyncReadyCallback   callback,
+								gpointer              data);
+static gboolean   g_daemon_file_input_stream_close_finish      (GInputStream         *stream,
+								GAsyncResult         *result,
+								GError              **error);
+static void       g_daemon_file_input_stream_query_info_async  (GFileInputStream     *stream,
+								const char           *attributes,
+								int                   io_priority,
+								GCancellable         *cancellable,
+								GAsyncReadyCallback   callback,
+								gpointer              user_data);
+static GFileInfo *g_daemon_file_input_stream_query_info_finish (GFileInputStream     *stream,
+								GAsyncResult         *result,
+								GError              **error);
+
 
 
 G_DEFINE_TYPE (GDaemonFileInputStream, g_daemon_file_input_stream,
 	       G_TYPE_FILE_INPUT_STREAM)
 
+static void
+pre_read_free (PreRead *pre)
+{
+  g_free (pre->data);
+  g_free (pre);
+}
+
+static void
+g_string_remove_in_front (GString *string,
+			  gsize bytes)
+{
+  memmove (string->str,
+	   string->str + bytes,
+	   string->len - bytes);
+  g_string_truncate (string,
+		     string->len - bytes);
+}
+
 static void
 g_daemon_file_input_stream_finalize (GObject *object)
 {
@@ -252,6 +315,14 @@ g_daemon_file_input_stream_finalize (GObject *object)
   if (file->data_stream)
     g_object_unref (file->data_stream);
 
+  while (file->pre_reads)
+    {
+      PreRead *pre = file->pre_reads->data;
+      file->pre_reads = g_list_delete_link (file->pre_reads,
+					    file->pre_reads);
+      pre_read_free (pre);
+    }
+  
   g_string_free (file->input_buffer, TRUE);
   g_string_free (file->output_buffer, TRUE);
   
@@ -286,6 +357,8 @@ g_daemon_file_input_stream_class_init (GDaemonFileInputStreamClass *klass)
   file_stream_class->can_seek = g_daemon_file_input_stream_can_seek;
   file_stream_class->seek = g_daemon_file_input_stream_seek;
   file_stream_class->query_info = g_daemon_file_input_stream_query_info;
+  file_stream_class->query_info_async = g_daemon_file_input_stream_query_info_async;
+  file_stream_class->query_info_finish = g_daemon_file_input_stream_query_info_finish;
 
 }
 
@@ -322,7 +395,8 @@ error_is_cancel (GError *error)
 
 static void
 append_request (GDaemonFileInputStream *stream, guint32 command,
-		guint32 arg1, guint32 arg2, guint32 *seq_nr)
+		guint32 arg1, guint32 arg2, guint32 data_len,
+		guint32 *seq_nr)
 {
   GVfsDaemonSocketProtocolRequest cmd;
 
@@ -332,10 +406,12 @@ append_request (GDaemonFileInputStream *stream, guint32 command,
     *seq_nr = stream->seq_nr;
   
   cmd.command = g_htonl (command);
-  cmd.seq_nr = g_htonl (stream->seq_nr++);
+  cmd.seq_nr = g_htonl (stream->seq_nr);
   cmd.arg1 = g_htonl (arg1);
   cmd.arg2 = g_htonl (arg2);
-  cmd.data_len = 0;
+  cmd.data_len = g_htonl (data_len);
+
+  stream->seq_nr++;
 
   g_string_append_len (stream->output_buffer,
 		       (char *)&cmd, G_VFS_DAEMON_SOCKET_PROTOCOL_REQUEST_SIZE);
@@ -356,7 +432,8 @@ get_reply_header_missing_bytes (GString *buffer)
   type = g_ntohl (reply->type);
   arg2 = g_ntohl (reply->arg2);
   
-  if (type == G_VFS_DAEMON_SOCKET_PROTOCOL_REPLY_ERROR)
+  if (type == G_VFS_DAEMON_SOCKET_PROTOCOL_REPLY_ERROR ||
+      type == G_VFS_DAEMON_SOCKET_PROTOCOL_REPLY_INFO)
     return G_VFS_DAEMON_SOCKET_PROTOCOL_REPLY_SIZE + arg2 - buffer->len;
   return 0;
 }
@@ -480,6 +557,7 @@ static StateOp
 iterate_read_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op, ReadOperation *op)
 {
   gsize len;
+  PreRead *pre;
 
   while (TRUE)
     {
@@ -487,6 +565,40 @@ iterate_read_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 	{
 	  /* Initial state for read op */
 	case READ_STATE_INIT:
+
+	  while (file->pre_reads)
+	    {
+	      pre = file->pre_reads->data;
+	      if (file->seek_generation != pre->seek_generation)
+		{
+		  file->pre_reads = g_list_delete_link (file->pre_reads,
+							file->pre_reads);
+		  pre_read_free (pre);
+		}
+	      else
+		{
+		  len = MIN (op->buffer_size, pre->len);
+		  memcpy (op->buffer, pre->data, len);
+		  op->ret_val = len;
+		  op->ret_error = NULL;
+
+		  if (len < pre->len)
+		    {
+		      memmove (pre->data, pre->data + len, pre->len - len);
+		      pre->len -= len;
+		    }
+		  else
+		    {
+		      file->pre_reads = g_list_delete_link (file->pre_reads,
+							    file->pre_reads);
+		      pre_read_free (pre);
+		    }
+		  
+		  return STATE_OP_DONE;
+		}
+	    }
+	  
+	  
 	  /* If we're already reading some data, but we didn't read all, just use that
 	     and don't even send a request */
 	  if (file->input_state == INPUT_STATE_IN_BLOCK &&
@@ -500,7 +612,7 @@ iterate_read_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 	    }
 
 	  append_request (file, G_VFS_DAEMON_SOCKET_PROTOCOL_REQUEST_READ,
-			  op->buffer_size, 0, &op->seq_nr);
+			  op->buffer_size, 0, 0, &op->seq_nr);
 	  op->state = READ_STATE_WROTE_COMMAND;
 	  io_op->io_buffer = file->output_buffer->str;
 	  io_op->io_size = file->output_buffer->len;
@@ -521,11 +633,8 @@ iterate_read_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 	  
 	  if (io_op->io_res < file->output_buffer->len)
 	    {
-	      memcpy (file->output_buffer->str,
-		      file->output_buffer->str + io_op->io_res,
-		      file->output_buffer->len - io_op->io_res);
-	      g_string_truncate (file->output_buffer,
-				 file->output_buffer->len - io_op->io_res);
+	      g_string_remove_in_front (file->output_buffer,
+					io_op->io_res);
 	      io_op->io_buffer = file->output_buffer->str;
 	      io_op->io_size = file->output_buffer->len;
 	      io_op->io_allow_cancel = FALSE;
@@ -542,7 +651,7 @@ iterate_read_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 	    {
 	      op->sent_cancel = TRUE;
 	      append_request (file, G_VFS_DAEMON_SOCKET_PROTOCOL_REQUEST_CANCEL,
-			      op->seq_nr, 0, NULL);
+			      op->seq_nr, 0, 0, NULL);
 	      op->state = READ_STATE_WROTE_COMMAND;
 	      io_op->io_buffer = file->output_buffer->str;
 	      io_op->io_size = file->output_buffer->len;
@@ -626,7 +735,7 @@ iterate_read_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 				 current_len + len);
 	      io_op->io_buffer = file->input_buffer->str + current_len;
 	      io_op->io_size = len;
-	      io_op->io_allow_cancel = !op->sent_cancel;
+	      io_op->io_allow_cancel = file->input_buffer->len == 0 && !op->sent_cancel;
 	      return STATE_OP_READ;
 	    }
 
@@ -763,8 +872,18 @@ iterate_close_state_machine (GDaemonFileInputStream *file, IOOperationData *io_o
 	{
 	  /* Initial state for read op */
 	case CLOSE_STATE_INIT:
+
+	  /* Clear any pre-read data blocks */
+	  while (file->pre_reads)
+	    {
+	      PreRead *pre = file->pre_reads->data;
+	      file->pre_reads = g_list_delete_link (file->pre_reads,
+						    file->pre_reads);
+	      pre_read_free (pre);
+	    }
+	  
 	  append_request (file, G_VFS_DAEMON_SOCKET_PROTOCOL_REQUEST_CLOSE,
-			  0, 0, &op->seq_nr);
+			  0, 0, 0, &op->seq_nr);
 	  op->state = CLOSE_STATE_WROTE_REQUEST;
 	  io_op->io_buffer = file->output_buffer->str;
 	  io_op->io_size = file->output_buffer->len;
@@ -785,11 +904,8 @@ iterate_close_state_machine (GDaemonFileInputStream *file, IOOperationData *io_o
 
 	  if (io_op->io_res < file->output_buffer->len)
 	    {
-	      memcpy (file->output_buffer->str,
-		      file->output_buffer->str + io_op->io_res,
-		      file->output_buffer->len - io_op->io_res);
-	      g_string_truncate (file->output_buffer,
-				 file->output_buffer->len - io_op->io_res);
+	      g_string_remove_in_front (file->output_buffer,
+					io_op->io_res);
 	      io_op->io_buffer = file->output_buffer->str;
 	      io_op->io_size = file->output_buffer->len;
 	      io_op->io_allow_cancel = FALSE;
@@ -806,7 +922,7 @@ iterate_close_state_machine (GDaemonFileInputStream *file, IOOperationData *io_o
 	    {
 	      op->sent_cancel = TRUE;
 	      append_request (file, G_VFS_DAEMON_SOCKET_PROTOCOL_REQUEST_CANCEL,
-			      op->seq_nr, 0, NULL);
+			      op->seq_nr, 0, 0, NULL);
 	      op->state = CLOSE_STATE_WROTE_REQUEST;
 	      io_op->io_buffer = file->output_buffer->str;
 	      io_op->io_size = file->output_buffer->len;
@@ -877,7 +993,7 @@ iterate_close_state_machine (GDaemonFileInputStream *file, IOOperationData *io_o
 				 current_len + len);
 	      io_op->io_buffer = file->input_buffer->str + current_len;
 	      io_op->io_size = len;
-	      io_op->io_allow_cancel = !op->sent_cancel;
+	      io_op->io_allow_cancel = file->input_buffer->len == 0 && !op->sent_cancel;
 	      return STATE_OP_READ;
 	    }
 
@@ -1013,6 +1129,7 @@ iterate_seek_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 	  append_request (file, request,
 			  op->offset & 0xffffffff,
 			  op->offset >> 32,
+			  0,
 			  &op->seq_nr);
 	  op->state = SEEK_STATE_WROTE_REQUEST;
 	  op->sent_seek = FALSE;
@@ -1039,13 +1156,19 @@ iterate_seek_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 	    file->seek_generation++;
 	  op->sent_seek = TRUE;
 	  
+	  /* Clear any pre-read data blocks */
+	  while (file->pre_reads)
+	    {
+	      PreRead *pre = file->pre_reads->data;
+	      file->pre_reads = g_list_delete_link (file->pre_reads,
+						    file->pre_reads);
+	      pre_read_free (pre);
+	    }
+	  
 	  if (io_op->io_res < file->output_buffer->len)
 	    {
-	      memcpy (file->output_buffer->str,
-		      file->output_buffer->str + io_op->io_res,
-		      file->output_buffer->len - io_op->io_res);
-	      g_string_truncate (file->output_buffer,
-				 file->output_buffer->len - io_op->io_res);
+	      g_string_remove_in_front (file->output_buffer,
+					io_op->io_res);
 	      io_op->io_buffer = file->output_buffer->str;
 	      io_op->io_size = file->output_buffer->len;
 	      io_op->io_allow_cancel = FALSE;
@@ -1062,7 +1185,7 @@ iterate_seek_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 	    {
 	      op->sent_cancel = TRUE;
 	      append_request (file, G_VFS_DAEMON_SOCKET_PROTOCOL_REQUEST_CANCEL,
-			      op->seq_nr, 0, NULL);
+			      op->seq_nr, 0, 0, NULL);
 	      op->state = SEEK_STATE_WROTE_REQUEST;
 	      io_op->io_buffer = file->output_buffer->str;
 	      io_op->io_size = file->output_buffer->len;
@@ -1134,7 +1257,7 @@ iterate_seek_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 				 current_len + len);
 	      io_op->io_buffer = file->input_buffer->str + current_len;
 	      io_op->io_size = len;
-	      io_op->io_allow_cancel = !op->sent_cancel;
+	      io_op->io_allow_cancel = file->input_buffer->len == 0 && !op->sent_cancel;
 	      return STATE_OP_READ;
 	    }
 
@@ -1228,17 +1351,276 @@ g_daemon_file_input_stream_seek (GFileInputStream *stream,
   return op.ret_val;
 }
 
+static StateOp
+iterate_query_state_machine (GDaemonFileInputStream *file,
+			     IOOperationData *io_op,
+			     QueryOperation *op)
+{
+  gsize len;
+  guint32 request;
+
+  while (TRUE)
+    {
+      switch (op->state)
+	{
+	  /* Initial state for read op */
+	case QUERY_STATE_INIT:
+	  request = G_VFS_DAEMON_SOCKET_PROTOCOL_REQUEST_QUERY_INFO;
+	  append_request (file, request,
+			  0,
+			  0,
+			  strlen (op->attributes),
+			  &op->seq_nr);
+	  g_string_append (file->output_buffer,
+			   op->attributes);
+	  
+	  op->state = QUERY_STATE_WROTE_REQUEST;
+	  io_op->io_buffer = file->output_buffer->str;
+	  io_op->io_size = file->output_buffer->len;
+	  io_op->io_allow_cancel = TRUE; /* Allow cancel before first byte of request sent */
+	  return STATE_OP_WRITE;
+
+	  /* wrote parts of output_buffer */
+	case QUERY_STATE_WROTE_REQUEST:
+	  if (io_op->io_cancelled)
+	    {
+	      op->info = NULL;
+	      g_set_error_literal (&op->ret_error,
+				   G_IO_ERROR,
+				   G_IO_ERROR_CANCELLED,
+				   _("Operation was cancelled"));
+	      return STATE_OP_DONE;
+	    }
+
+	  if (io_op->io_res < file->output_buffer->len)
+	    {
+	      g_string_remove_in_front (file->output_buffer,
+					io_op->io_res);
+	      io_op->io_buffer = file->output_buffer->str;
+	      io_op->io_size = file->output_buffer->len;
+	      io_op->io_allow_cancel = FALSE;
+	      return STATE_OP_WRITE;
+	    }
+	  g_string_truncate (file->output_buffer, 0);
+
+	  op->state = QUERY_STATE_HANDLE_INPUT;
+	  break;
+
+	  /* No op */
+	case QUERY_STATE_HANDLE_INPUT:
+	  if (io_op->cancelled && !op->sent_cancel)
+	    {
+	      op->sent_cancel = TRUE;
+	      append_request (file, G_VFS_DAEMON_SOCKET_PROTOCOL_REQUEST_CANCEL,
+			      op->seq_nr, 0, 0, NULL);
+	      op->state = QUERY_STATE_WROTE_REQUEST;
+	      io_op->io_buffer = file->output_buffer->str;
+	      io_op->io_size = file->output_buffer->len;
+	      io_op->io_allow_cancel = FALSE;
+	      return STATE_OP_WRITE;
+	    }
+	  
+	  if (file->input_state == INPUT_STATE_IN_BLOCK)
+	    {
+	      op->state = QUERY_STATE_HANDLE_INPUT_BLOCK;
+	      break;
+	    }
+	  else if (file->input_state == INPUT_STATE_IN_REPLY_HEADER)
+	    {
+	      op->state = QUERY_STATE_HANDLE_HEADER;
+	      break;
+	    }
+	  g_assert_not_reached ();
+	  break;
+
+	  /* No op */
+	case QUERY_STATE_HANDLE_INPUT_BLOCK:
+	  g_assert (file->input_state == INPUT_STATE_IN_BLOCK);
+
+	  if (file->input_block_size == 0)
+	    {
+	      file->input_state = INPUT_STATE_IN_REPLY_HEADER;
+	      op->state = QUERY_STATE_HANDLE_INPUT;
+	      break;
+	    }
+	  
+	  if (file->seek_generation ==
+	      file->input_block_seek_generation)
+	    {
+	      op->state = QUERY_STATE_READ_BLOCK;
+	      io_op->io_buffer = g_malloc (file->input_block_size); 
+	      io_op->io_size = file->input_block_size;
+	      io_op->io_allow_cancel = FALSE;
+	      return STATE_OP_READ;
+	    }
+	  else
+	    {
+	      op->state = QUERY_STATE_SKIP_BLOCK;
+	      io_op->io_buffer = NULL;
+	      io_op->io_size = file->input_block_size;
+	      io_op->io_allow_cancel = !op->sent_cancel;
+	      return STATE_OP_SKIP;
+	    }
+	  break;
+
+	  /* Read block data */
+	case QUERY_STATE_SKIP_BLOCK:
+	  if (io_op->io_cancelled)
+	    {
+	      op->state = QUERY_STATE_HANDLE_INPUT;
+	      break;
+	    }
+	  
+	  g_assert (io_op->io_res <= file->input_block_size);
+	  file->input_block_size -= io_op->io_res;
+	  
+	  if (file->input_block_size == 0)
+	    file->input_state = INPUT_STATE_IN_REPLY_HEADER;
+	  
+	  op->state = QUERY_STATE_HANDLE_INPUT;
+	  break;
+
+	  /* Read block data */
+	case QUERY_STATE_READ_BLOCK:
+	  if (io_op->io_cancelled)
+	    {
+	      g_free (io_op->io_buffer);
+	      op->state = QUERY_STATE_HANDLE_INPUT;
+	      break;
+	    }
+	  
+	  if (io_op->io_res > 0)
+	    {
+	      PreRead *pre;
+	      
+	      g_assert (io_op->io_res <= file->input_block_size);
+	      file->input_block_size -= io_op->io_res;
+	      if (file->input_block_size == 0)
+		file->input_state = INPUT_STATE_IN_REPLY_HEADER;
+
+	      pre = g_new (PreRead, 1);
+	      pre->data = io_op->io_buffer;
+	      pre->len = io_op->io_res;
+	      pre->seek_generation = file->input_block_seek_generation;
+
+	      file->pre_reads = g_list_append (file->pre_reads, pre);
+	    }
+	  else
+	    g_free (io_op->io_buffer);
+	  
+	  op->state = QUERY_STATE_HANDLE_INPUT;
+	  break;
+	  
+	  /* read header data, (or manual io_len/res = 0) */
+	case QUERY_STATE_HANDLE_HEADER:
+	  if (io_op->io_cancelled)
+	    {
+	      op->state = QUERY_STATE_HANDLE_INPUT;
+	      break;
+	    }
+
+	  if (io_op->io_res > 0)
+	    {
+	      gsize unread_size = io_op->io_size - io_op->io_res;
+	      g_string_set_size (file->input_buffer,
+				 file->input_buffer->len - unread_size);
+	    }
+	  
+	  len = get_reply_header_missing_bytes (file->input_buffer);
+	  if (len > 0)
+	    {
+	      gsize current_len = file->input_buffer->len;
+	      g_string_set_size (file->input_buffer,
+				 current_len + len);
+	      io_op->io_buffer = file->input_buffer->str + current_len;
+	      io_op->io_size = len;
+	      io_op->io_allow_cancel = file->input_buffer->len == 0 && !op->sent_cancel;
+	      return STATE_OP_READ;
+	    }
+
+	  /* Got full header */
+
+	  {
+	    GVfsDaemonSocketProtocolReply reply;
+	    char *data;
+	    data = decode_reply (file->input_buffer, &reply);
+
+	    if (reply.type == G_VFS_DAEMON_SOCKET_PROTOCOL_REPLY_ERROR &&
+		reply.seq_nr == op->seq_nr)
+	      {
+		op->info = NULL;
+		decode_error (&reply, data, &op->ret_error);
+		g_string_truncate (file->input_buffer, 0);
+		return STATE_OP_DONE;
+	      }
+	    else if (reply.type == G_VFS_DAEMON_SOCKET_PROTOCOL_REPLY_DATA)
+	      {
+		g_string_truncate (file->input_buffer, 0);
+		file->input_state = INPUT_STATE_IN_BLOCK;
+		file->input_block_size = reply.arg1;
+		file->input_block_seek_generation = reply.arg2;
+		op->state = QUERY_STATE_HANDLE_INPUT_BLOCK;
+		break;
+	      }
+	    else if (reply.type == G_VFS_DAEMON_SOCKET_PROTOCOL_REPLY_INFO)
+	      {
+		op->info = gvfs_file_info_demarshal (data, reply.arg2);
+		g_string_truncate (file->input_buffer, 0);
+		return STATE_OP_DONE;
+	      }
+	    /* Ignore other reply types */
+	  }
+
+	  g_string_truncate (file->input_buffer, 0);
+	  
+	  /* This wasn't interesting, read next reply */
+	  op->state = QUERY_STATE_HANDLE_HEADER;
+	  break;
+
+	default:
+	  g_assert_not_reached ();
+	}
+      
+      /* Clear io_op between non-op state switches */
+      io_op->io_size = 0;
+      io_op->io_res = 0;
+      io_op->io_cancelled = FALSE;
+    }
+}
+
+
 static GFileInfo *
 g_daemon_file_input_stream_query_info (GFileInputStream     *stream,
-				       char                 *attributes,
+				       const char           *attributes,
 				       GCancellable         *cancellable,
 				       GError              **error)
 {
-  g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED, _("The query info operation is not supported"));
+   GDaemonFileInputStream *file;
+   QueryOperation op;
+
+  file = G_DAEMON_FILE_INPUT_STREAM (stream);
+
+  if (g_cancellable_set_error_if_cancelled (cancellable, error))
+    return NULL;
   
-  return NULL;
+  memset (&op, 0, sizeof (op));
+  op.state = QUERY_STATE_INIT;
+  if (attributes)
+    op.attributes = (char *)attributes;
+  else
+    op.attributes = "";
+  
+  if (!run_sync_state_machine (file, (state_machine_iterator)iterate_query_state_machine,
+			       &op, cancellable, error))
+    return NULL; /* IO Error */
+
+  if (op.info == NULL)
+    g_propagate_error (error, op.ret_error);
+  
+  return op.info;
 }
 
+
 /************************************************************************
  *         Async I/O Code                                               *
  ************************************************************************/
@@ -1641,3 +2023,96 @@ g_daemon_file_input_stream_close_finish (GInputStream              *stream,
   /* Failures handled in generic close_finish code */
   return TRUE;
 }
+
+static void
+async_query_done (GInputStream *stream,
+		  gpointer op_data,
+		  GAsyncReadyCallback callback,
+		  gpointer user_data,
+		  GError *io_error)
+{
+  GDaemonFileInputStream *file;
+  GSimpleAsyncResult *simple;
+  QueryOperation *op;
+  GFileInfo *info;
+  GError *error;
+
+  file = G_DAEMON_FILE_INPUT_STREAM (stream);
+  
+  op = op_data;
+
+  if (io_error)
+    {
+      info = NULL;
+      error = io_error;
+    }
+  else
+    {
+      info = op->info;
+      error = op->ret_error;
+    }
+
+  simple = g_simple_async_result_new (G_OBJECT (stream),
+				      callback, user_data,
+				      g_daemon_file_input_stream_query_info_async);
+
+  if (info == NULL)
+    g_simple_async_result_set_from_error (simple, error);
+  else
+    g_simple_async_result_set_op_res_gpointer (simple, info,
+					       g_object_unref);
+  
+  /* Complete immediately, not in idle, since we're already in a mainloop callout */
+  g_simple_async_result_complete (simple);
+  g_object_unref (simple);
+  
+  if (op->ret_error)
+    g_error_free (op->ret_error);
+  g_free (op->attributes);
+  g_free (op);
+}
+
+static void
+g_daemon_file_input_stream_query_info_async  (GFileInputStream     *stream,
+					      const char           *attributes,
+					      int                   io_priority,
+					      GCancellable         *cancellable,
+					      GAsyncReadyCallback   callback,
+					      gpointer              user_data)
+{
+  GDaemonFileInputStream *file;
+  QueryOperation *op;
+
+  file = G_DAEMON_FILE_INPUT_STREAM (stream);
+  
+  op = g_new0 (QueryOperation, 1);
+  op->state = QUERY_STATE_INIT;
+  if (attributes)
+    op->attributes = g_strdup (attributes);
+  else
+    op->attributes = g_strdup ("");
+
+  run_async_state_machine (file,
+			   (state_machine_iterator)iterate_query_state_machine,
+			   op, io_priority,
+			   callback, user_data,
+			   cancellable,
+			   async_query_done);
+}
+
+static GFileInfo *
+g_daemon_file_input_stream_query_info_finish (GFileInputStream     *stream,
+					      GAsyncResult         *result,
+					      GError              **error)
+{
+  GSimpleAsyncResult *simple;
+  GFileInfo *info;
+
+  simple = G_SIMPLE_ASYNC_RESULT (result);
+  g_assert (g_simple_async_result_get_source_tag (simple) == g_daemon_file_input_stream_query_info_async);
+  
+  info = g_simple_async_result_get_op_res_gpointer (simple);
+  
+  return g_object_ref (info);
+  
+}
