@@ -43,32 +43,36 @@ typedef struct _GVfsBackend        GVfsBackend;
 typedef struct _GVfsBackendPrivate GVfsBackendPrivate;
 typedef struct _GVfsBackendClass   GVfsBackendClass;
 
-typedef struct _GVfsJobMount           GVfsJobMount;
-typedef struct _GVfsJobUnmount         GVfsJobUnmount;
-typedef struct _GVfsJobMountMountable  GVfsJobMountMountable;
+typedef struct _GVfsJobMount            GVfsJobMount;
+typedef struct _GVfsJobUnmount          GVfsJobUnmount;
+typedef struct _GVfsJobMountMountable   GVfsJobMountMountable;
 typedef struct _GVfsJobUnmountMountable GVfsJobUnmountMountable;
-typedef struct _GVfsJobOpenForRead     GVfsJobOpenForRead;
-typedef struct _GVfsJobSeekRead        GVfsJobSeekRead;
-typedef struct _GVfsJobCloseRead       GVfsJobCloseRead;
-typedef struct _GVfsJobRead            GVfsJobRead;
-typedef struct _GVfsJobOpenForWrite    GVfsJobOpenForWrite;
-typedef struct _GVfsJobWrite           GVfsJobWrite;
-typedef struct _GVfsJobSeekWrite       GVfsJobSeekWrite;
-typedef struct _GVfsJobCloseWrite      GVfsJobCloseWrite;
-typedef struct _GVfsJobQueryInfo       GVfsJobQueryInfo;
-typedef struct _GVfsJobQueryFsInfo     GVfsJobQueryFsInfo;
-typedef struct _GVfsJobEnumerate       GVfsJobEnumerate;
-typedef struct _GVfsJobSetDisplayName  GVfsJobSetDisplayName;
-typedef struct _GVfsJobTrash           GVfsJobTrash;
-typedef struct _GVfsJobDelete          GVfsJobDelete;
-typedef struct _GVfsJobMakeDirectory   GVfsJobMakeDirectory;
-typedef struct _GVfsJobMakeSymlink     GVfsJobMakeSymlink;
-typedef struct _GVfsJobCopy            GVfsJobCopy;
-typedef struct _GVfsJobUpload          GVfsJobUpload;
-typedef struct _GVfsJobMove            GVfsJobMove;
-typedef struct _GVfsJobSetAttribute    GVfsJobSetAttribute;
-typedef struct _GVfsJobQueryAttributes GVfsJobQueryAttributes;
-typedef struct _GVfsJobCreateMonitor   GVfsJobCreateMonitor;
+typedef struct _GVfsJobOpenForRead      GVfsJobOpenForRead;
+typedef struct _GVfsJobOpenIconForRead  GVfsJobOpenIconForRead;
+typedef struct _GVfsJobSeekRead         GVfsJobSeekRead;
+typedef struct _GVfsJobCloseRead        GVfsJobCloseRead;
+typedef struct _GVfsJobRead             GVfsJobRead;
+typedef struct _GVfsJobOpenForWrite     GVfsJobOpenForWrite;
+typedef struct _GVfsJobWrite            GVfsJobWrite;
+typedef struct _GVfsJobSeekWrite        GVfsJobSeekWrite;
+typedef struct _GVfsJobCloseWrite       GVfsJobCloseWrite;
+typedef struct _GVfsJobQueryInfo        GVfsJobQueryInfo;
+typedef struct _GVfsJobQueryInfoRead    GVfsJobQueryInfoRead;
+typedef struct _GVfsJobQueryInfoWrite   GVfsJobQueryInfoWrite;
+typedef struct _GVfsJobQueryFsInfo      GVfsJobQueryFsInfo;
+typedef struct _GVfsJobEnumerate        GVfsJobEnumerate;
+typedef struct _GVfsJobSetDisplayName   GVfsJobSetDisplayName;
+typedef struct _GVfsJobTrash            GVfsJobTrash;
+typedef struct _GVfsJobDelete           GVfsJobDelete;
+typedef struct _GVfsJobMakeDirectory    GVfsJobMakeDirectory;
+typedef struct _GVfsJobMakeSymlink      GVfsJobMakeSymlink;
+typedef struct _GVfsJobCopy             GVfsJobCopy;
+typedef struct _GVfsJobMove             GVfsJobMove;
+typedef struct _GVfsJobPush             GVfsJobPush;
+typedef struct _GVfsJobPull             GVfsJobPull;
+typedef struct _GVfsJobSetAttribute     GVfsJobSetAttribute;
+typedef struct _GVfsJobQueryAttributes  GVfsJobQueryAttributes;
+typedef struct _GVfsJobCreateMonitor    GVfsJobCreateMonitor;
 
 typedef gpointer GVfsBackendHandle;
 
@@ -137,6 +141,12 @@ struct _GVfsBackendClass
   gboolean (*try_open_for_read) (GVfsBackend *backend,
 				 GVfsJobOpenForRead *job,
 				 const char *filename);
+  void     (*open_icon_for_read) (GVfsBackend *backend,
+                                  GVfsJobOpenIconForRead *job,
+                                  const char *icon_id);
+  gboolean (*try_open_icon_for_read) (GVfsBackend *backend,
+                                      GVfsJobOpenIconForRead *job,
+                                      const char *icon_id);
   void     (*close_read)        (GVfsBackend *backend,
 				 GVfsJobCloseRead *job,
 				 GVfsBackendHandle handle);
@@ -229,6 +239,26 @@ struct _GVfsBackendClass
 				 GFileQueryInfoFlags flags,
 				 GFileInfo *info,
 				 GFileAttributeMatcher *attribute_matcher);
+  void     (*query_info_on_read)(GVfsBackend *backend,
+				 GVfsJobQueryInfoRead *job,
+				 GVfsBackendHandle handle,
+				 GFileInfo *info,
+				 GFileAttributeMatcher *attribute_matcher);
+  gboolean (*try_query_info_on_read)(GVfsBackend *backend,
+				 GVfsJobQueryInfoRead *job,
+				 GVfsBackendHandle handle,
+				 GFileInfo *info,
+				 GFileAttributeMatcher *attribute_matcher);
+  void     (*query_info_on_write)(GVfsBackend *backend,
+				 GVfsJobQueryInfoWrite *job,
+				 GVfsBackendHandle handle,
+				 GFileInfo *info,
+				 GFileAttributeMatcher *attribute_matcher);
+  gboolean (*try_query_info_on_write)(GVfsBackend *backend,
+				 GVfsJobQueryInfoWrite *job,
+				 GVfsBackendHandle handle,
+				 GFileInfo *info,
+				 GFileAttributeMatcher *attribute_matcher);
   void     (*query_fs_info)     (GVfsBackend *backend,
 				 GVfsJobQueryFsInfo *job,
 				 const char *filename,
@@ -297,21 +327,7 @@ struct _GVfsBackendClass
 				 GFileCopyFlags flags,
 				 GFileProgressCallback progress_callback,
 				 gpointer progress_callback_data);
-  void     (*upload)            (GVfsBackend *backend,
-				 GVfsJobUpload *job,
-				 const char *destination,
-				 const char *local_path,
-				 GFileCopyFlags flags,
-				 GFileProgressCallback progress_callback,
-				 gpointer progress_callback_data);
-  gboolean (*try_upload)        (GVfsBackend *backend,
-				 GVfsJobUpload *job,
-				 const char *destination,
-				 const char *local_path,
-				 GFileCopyFlags flags,
-				 GFileProgressCallback progress_callback,
-				 gpointer progress_callback_data);
-  void     (*move)              (GVfsBackend *backend,
+   void     (*move)              (GVfsBackend *backend,
 				 GVfsJobMove *job,
 				 const char *source,
 				 const char *destination,
@@ -325,6 +341,38 @@ struct _GVfsBackendClass
 				 GFileCopyFlags flags,
 				 GFileProgressCallback progress_callback,
 				 gpointer progress_callback_data);
+  void     (*push)             (GVfsBackend *backend,
+                                GVfsJobPush *job,
+                                const char *destination,
+                                const char *local_path,
+                                GFileCopyFlags flags,
+                                gboolean remove_source,
+                                GFileProgressCallback progress_callback,
+                                gpointer progress_callback_data);
+  gboolean (*try_push)         (GVfsBackend *backend,
+                                GVfsJobPush *job,
+                                const char *destination,
+                                const char *local_path,
+                                GFileCopyFlags flags,
+                                gboolean remove_source,
+                                GFileProgressCallback progress_callback,
+                                gpointer progress_callback_data);
+  void     (*pull)             (GVfsBackend *backend,
+                                GVfsJobPull *job,
+                                const char *source,
+                                const char *local_path,
+                                GFileCopyFlags flags,
+                                gboolean remove_source,
+                                GFileProgressCallback progress_callback,
+                                gpointer progress_callback_data);
+  gboolean (*try_pull)         (GVfsBackend *backend,
+                                GVfsJobPull *job,
+                                const char *source,
+                                const char *local_path,
+                                GFileCopyFlags flags,
+                                gboolean remove_source,
+                                GFileProgressCallback progress_callback,
+                                gpointer progress_callback_data);
   void     (*set_attribute)     (GVfsBackend *backend,
 				 GVfsJobSetAttribute *set_attribute,
 				 const char *filename,
@@ -382,7 +430,9 @@ void        g_vfs_backend_set_stable_name                (GVfsBackend        *ba
 void        g_vfs_backend_set_x_content_types            (GVfsBackend        *backend,
 							  char              **x_content_types);
 void        g_vfs_backend_set_icon_name                  (GVfsBackend        *backend,
-							  const char         *icon);
+							  const char         *icon_name);
+void        g_vfs_backend_set_icon                       (GVfsBackend        *backend,
+							  GIcon              *icon);
 void        g_vfs_backend_set_prefered_filename_encoding (GVfsBackend        *backend,
 							  const char         *prefered_filename_encoding);
 void        g_vfs_backend_set_user_visible               (GVfsBackend        *backend,
@@ -399,7 +449,7 @@ const char *g_vfs_backend_get_backend_type               (GVfsBackend        *ba
 const char *g_vfs_backend_get_display_name               (GVfsBackend        *backend);
 const char *g_vfs_backend_get_stable_name                (GVfsBackend        *backend);
 char      **g_vfs_backend_get_x_content_types            (GVfsBackend        *backend);
-const char *g_vfs_backend_get_icon_name                  (GVfsBackend        *backend);
+GIcon      *g_vfs_backend_get_icon                       (GVfsBackend        *backend);
 GMountSpec *g_vfs_backend_get_mount_spec                 (GVfsBackend        *backend);
 GVfsDaemon *g_vfs_backend_get_daemon                     (GVfsBackend        *backend);
 
