@@ -17,7 +17,8 @@
  * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  * Boston, MA 02111-1307, USA.
  *
- * Author: Bastien Nocera <hadess@hadess.net>
+ * Authors: Bastien Nocera <hadess@hadess.net>
+ *          Cosimo Cecchi  <cosimoc@gnome.org>
  */
 
 #include <config.h>
@@ -28,11 +29,15 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <string.h>
+#include <stdlib.h>
 
 #include <glib/gstdio.h>
 #include <glib/gi18n.h>
 #include <gio/gio.h>
+#include <libhal.h>
 #include <dbus/dbus-glib.h>
+#include <dbus/dbus-glib-bindings.h>
+#include <dbus/dbus-glib-lowlevel.h>
 #include <bluetooth/bluetooth.h>
 
 #include "gvfsbackendobexftp.h"
@@ -57,7 +62,8 @@
 
 #define BDADDR_LEN 17
 
-#define ASYNC_SUCCESS 1
+#define ASYNC_SUCCESS 2
+#define ASYNC_RUNNING 1
 #define ASYNC_PENDING 0
 #define ASYNC_ERROR -1
 
@@ -70,6 +76,7 @@ struct _GVfsBackendObexftp
   char *display_name;
   char *bdaddr;
   char *icon_name;
+  gint usbintfnum;
 
   DBusGConnection *connection;
   DBusGProxy *manager_proxy;
@@ -105,18 +112,70 @@ static void session_connected_cb (DBusGProxy *proxy,
                                   const char *session_object,
                                   gpointer user_data);
 
+/* Parse USB paths from do_mount(), or obex-data-server */
+static gboolean
+_get_numbers_from_usb_path (const char *path, int *usb_bus_num, int *usb_device_num, int *usb_intf_num)
+{
+  char **tokens;
+  char *endp;
+  gboolean has_brackets = FALSE;
+
+  if (path == NULL)
+    return FALSE;
+  if (*path == '[')
+    {
+      path++;
+      has_brackets = TRUE;
+    }
+
+  tokens = g_strsplit (path + 4, ",", 0);
+  if (g_strv_length (tokens) != 3)
+    {
+      g_strfreev (tokens);
+      return FALSE;
+    }
+
+ *usb_bus_num = strtol (tokens[0], &endp, 10);
+  if (*endp != '\0')
+    {
+      g_strfreev (tokens);
+      return FALSE;
+    }
+
+  *usb_device_num = strtol (tokens[1], &endp, 10);
+  if (*endp != '\0')
+    {
+      g_strfreev (tokens);
+      return FALSE;
+    }
+
+  *usb_intf_num = strtol (tokens[2], &endp, 10);
+  if ((has_brackets && *endp != ']') || (!has_brackets && *endp != '\0'))
+    {
+      g_strfreev (tokens);
+      return FALSE;
+    }
+
+  g_strfreev (tokens);
+
+  return TRUE;
+}
+
 /* Used to detect broken listings from
  * old Nokia 3650s */
 static gboolean
 _is_nokia_3650 (const char *bdaddr)
 {
+  if (!bdaddr)
+    return FALSE;
+
   /* Don't ask, Nokia seem to use a Bluetooth
    * HCI from Murata */
   return g_str_has_prefix(bdaddr, "00:60:57");
 }
 
 static char *
-get_name_and_icon (DBusGProxy *device, char **icon_name)
+_get_bluetooth_name_and_icon (DBusGProxy *device, char **icon_name)
 {
   GHashTable *hash;
 
@@ -146,10 +205,14 @@ get_name_and_icon (DBusGProxy *device, char **icon_name)
   return NULL;
 }
 
+
+#define _DBUS_POINTER_SHIFT(p)   ((void*) (((char*)p) + sizeof (void*)))
+#define DBUS_G_CONNECTION_FROM_CONNECTION(x)     ((DBusGConnection*) _DBUS_POINTER_SHIFT(x))
+
 static gchar *
-_get_device_properties (const char *bdaddr, char **icon_name)
+_get_bluetooth_device_properties (const char *bdaddr, char **icon_name)
 {
-  DBusGConnection *connection;
+  DBusConnection *conn;
   DBusGProxy *manager;
   GPtrArray *adapters;
   gchar *name;
@@ -157,22 +220,24 @@ _get_device_properties (const char *bdaddr, char **icon_name)
 
   name = NULL;
 
-  connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, NULL);
-  if (connection == NULL)
+  conn = dbus_bus_get_private (DBUS_BUS_SYSTEM, NULL);
+  if (conn == NULL)
         return name;
 
-  manager = dbus_g_proxy_new_for_name (connection, "org.bluez",
+  manager = dbus_g_proxy_new_for_name (DBUS_G_CONNECTION_FROM_CONNECTION(conn), "org.bluez",
                                        "/", "org.bluez.Manager");
   if (manager == NULL)
     {
-      dbus_g_connection_unref (connection);
+      dbus_connection_close (conn);
+      dbus_connection_unref (conn);
       return name;
     }
 
   if (dbus_g_proxy_call (manager, "ListAdapters", NULL, G_TYPE_INVALID, dbus_g_type_get_collection ("GPtrArray", DBUS_TYPE_G_OBJECT_PATH), &adapters, G_TYPE_INVALID) == FALSE)
     {
       g_object_unref (manager);
-      dbus_g_connection_unref (connection);
+      dbus_connection_close (conn);
+      dbus_connection_unref (conn);
       return name;
     }
 
@@ -181,15 +246,15 @@ _get_device_properties (const char *bdaddr, char **icon_name)
       DBusGProxy *adapter;
       char *device_path;
 
-      adapter = dbus_g_proxy_new_for_name (connection, "org.bluez",
+      adapter = dbus_g_proxy_new_for_name (DBUS_G_CONNECTION_FROM_CONNECTION(conn), "org.bluez",
                                            g_ptr_array_index (adapters, i), "org.bluez.Adapter");
       if (dbus_g_proxy_call (adapter, "FindDevice", NULL,
                              G_TYPE_STRING, bdaddr, G_TYPE_INVALID,
                              DBUS_TYPE_G_OBJECT_PATH, &device_path, G_TYPE_INVALID) != FALSE)
         {
           DBusGProxy *device;
-          device = dbus_g_proxy_new_for_name (connection, "org.bluez", device_path, "org.bluez.Device");
-          name = get_name_and_icon (device, icon_name);
+          device = dbus_g_proxy_new_for_name (DBUS_G_CONNECTION_FROM_CONNECTION(conn), "org.bluez", device_path, "org.bluez.Device");
+          name = _get_bluetooth_name_and_icon (device, icon_name);
           g_object_unref (device);
         }
       g_object_unref (adapter);
@@ -197,11 +262,215 @@ _get_device_properties (const char *bdaddr, char **icon_name)
 
   g_ptr_array_free (adapters, TRUE);
   g_object_unref (manager);
-  dbus_g_connection_unref (connection);
+  dbus_connection_close (conn);
+  dbus_connection_unref (conn);
 
   return name;
 }
 
+static gboolean
+_is_same_path (const char *path, int usb_bus_num, int usb_device_num, int usb_intf_num)
+{
+  int bus, dev, intf;
+
+  if (!_get_numbers_from_usb_path (path, &bus, &dev, &intf))
+    return FALSE;
+
+  if (bus == usb_bus_num &&
+      dev == usb_device_num &&
+      intf == usb_intf_num)
+        return TRUE;
+
+  return FALSE;
+}
+
+static int
+_find_ods_usb_intfnum (DBusGProxy *obex_manager, int device_usb_bus_num, int device_usb_device_num, int device_usb_interface_num)
+{
+  int i, n;
+  GHashTable *hash;
+
+  if (obex_manager == NULL)
+    return -1;
+
+  if (dbus_g_proxy_call (obex_manager, "GetUsbInterfacesNum", NULL, G_TYPE_INVALID, G_TYPE_UINT, &n, G_TYPE_INVALID) == FALSE)
+    return -1;
+
+  for (i = 0; i < n; i++) 
+    {
+      if (dbus_g_proxy_call (obex_manager, "GetUsbInterfaceInfo", NULL,
+                             G_TYPE_UINT, i, G_TYPE_INVALID,
+                             DBUS_TYPE_G_STRING_STRING_HASHTABLE, &hash, 
+                             G_TYPE_INVALID) != FALSE)
+        {
+          const char* ods_path = g_hash_table_lookup (hash, "Path");
+
+          if (ods_path == NULL)
+            {
+              g_hash_table_destroy (hash);
+              return -2;
+            }
+
+          if (_is_same_path (ods_path, device_usb_bus_num, device_usb_device_num, device_usb_interface_num))
+            {
+              g_hash_table_destroy (hash);
+              return i;
+            }
+          g_hash_table_destroy (hash);
+        }
+    }
+  return -1;
+}
+
+static gint
+_get_usb_intfnum_and_properties (DBusGProxy *obex_manager, const char *device, char **display_name, char **icon_name)
+{
+  char **obex_devices;
+  int num_obex_devices;
+  int usb_bus_num;
+  int usb_device_num;
+  int usb_intf_num;
+  DBusError dbus_error;
+  DBusConnection *dbus_connection;
+  LibHalContext *hal_ctx;
+  int ods_intf_num = 1;
+  int n;
+
+  /* Parse the [usb:1,41,3] string */
+  if (!g_str_has_prefix (device, "[usb:"))
+    {
+      return -1;
+    }
+
+  if (!_get_numbers_from_usb_path (device, &usb_bus_num, &usb_device_num, &usb_intf_num))
+    return -1;
+
+  g_message ("Parsed '%s' into bus=%d device=%d interface=%d", device, usb_bus_num, usb_device_num, usb_intf_num);
+
+  dbus_error_init (&dbus_error);
+  dbus_connection = dbus_bus_get_private (DBUS_BUS_SYSTEM, &dbus_error);
+  if (dbus_error_is_set (&dbus_error))
+    {
+      dbus_error_free (&dbus_error);
+      return -1;
+    }
+  dbus_connection_set_exit_on_disconnect (dbus_connection, FALSE);
+
+  hal_ctx = libhal_ctx_new ();
+  if (hal_ctx == NULL)
+    {
+      dbus_connection_close (dbus_connection);
+      dbus_connection_unref (dbus_connection);
+      dbus_error_free (&dbus_error);
+      return -1;
+    }
+
+  libhal_ctx_set_dbus_connection (hal_ctx, dbus_connection);
+  if (!libhal_ctx_init (hal_ctx, &dbus_error))
+    {
+      libhal_ctx_free (hal_ctx);
+      dbus_connection_close (dbus_connection);
+      dbus_connection_unref (dbus_connection);
+      dbus_error_free (&dbus_error);
+      return -1;
+    }
+  obex_devices = libhal_find_device_by_capability (hal_ctx, "obex", &num_obex_devices, NULL);
+
+  for (n = 0; n < num_obex_devices; n++)
+    {
+      char *udi = obex_devices[n];
+      LibHalPropertySet *ps;
+
+      ps = libhal_device_get_all_properties (hal_ctx, udi, NULL);
+      if (ps != NULL)
+        {
+          const char *subsystem;
+
+          subsystem = libhal_ps_get_string (ps, "info.subsystem");
+          if (subsystem != NULL && strcmp (subsystem, "usb") == 0)
+            {
+              int device_usb_bus_num;
+              int device_usb_device_num;
+              int device_usb_interface_num;
+              const char *icon_from_hal;
+              const char *name_from_hal;
+
+              device_usb_bus_num = libhal_ps_get_int32 (ps, "usb.bus_number");
+              device_usb_device_num = libhal_ps_get_int32 (ps, "usb.linux.device_number");
+              device_usb_interface_num = libhal_ps_get_int32 (ps, "usb.interface.number");
+              icon_from_hal = libhal_ps_get_string (ps, "info.desktop.icon");
+              name_from_hal = libhal_ps_get_string (ps, "info.desktop.name");
+
+              g_message ("looking at usb device '%s' with bus=%d, device=%d interface=%d", 
+                     udi, device_usb_bus_num, device_usb_device_num, device_usb_interface_num);
+
+              if (device_usb_bus_num == usb_bus_num &&
+                  device_usb_device_num == usb_device_num &&
+                  device_usb_interface_num == usb_intf_num)
+                {
+                  const char *parent_udi;
+
+                  g_message ("udi '%s' matches %s", udi, device);
+                  parent_udi = libhal_ps_get_string (ps, "info.parent");
+
+                  if (parent_udi != NULL)
+                    {
+                      LibHalPropertySet *ps2;
+
+                      ps2 = libhal_device_get_all_properties (hal_ctx, parent_udi, NULL);
+                      if (ps2 != NULL)
+                        {
+                          ods_intf_num = _find_ods_usb_intfnum (obex_manager,
+                                                                device_usb_bus_num,
+                                                                device_usb_device_num,
+                                                                device_usb_interface_num);
+
+                          if (ods_intf_num >= 0)
+                            {
+                              if (icon_from_hal != NULL)
+                                *icon_name = g_strdup (icon_from_hal);
+                              else
+                                *icon_name = "drive-removable-media-usb";
+
+                              if (name_from_hal != NULL)
+                                *display_name = g_strdup (name_from_hal);
+                              else
+                                *display_name = g_strdup_printf ("%s - %s",
+                                                                 libhal_ps_get_string (ps2, "usb_device.vendor"),
+                                                                 libhal_ps_get_string (ps2, "usb_device.product"));
+
+                              libhal_free_property_set (ps2);
+                              libhal_free_property_set (ps);
+
+                              goto end;
+                            }
+                          else if (ods_intf_num == -2)
+                            {
+                              libhal_free_property_set (ps2);
+                              libhal_free_property_set (ps);
+
+                              goto end;
+                            }
+                          libhal_free_property_set (ps2);
+                        }
+                    }
+                }
+            }
+          libhal_free_property_set (ps);
+        }
+    }
+
+end:
+  libhal_free_string_array (obex_devices);
+  libhal_ctx_free (hal_ctx);
+
+  dbus_connection_close (dbus_connection);
+  dbus_connection_unref (dbus_connection);
+  dbus_error_free (&dbus_error);
+
+  return ods_intf_num;
+}
+
 static void
 g_vfs_backend_obexftp_finalize (GObject *object)
 {
@@ -365,7 +634,7 @@ _query_file_info_helper (GVfsBackend *backend,
   GList *elements, *l;
   gboolean found;
 
-  g_print ("+ _query_file_info_helper, filename: %s\n", filename);
+  g_debug ("+ _query_file_info_helper, filename: %s\n", filename);
 
   if (strcmp (filename, "/") == 0)
     {
@@ -376,7 +645,14 @@ _query_file_info_helper (GVfsBackend *backend,
       g_file_info_set_file_type (info, G_FILE_TYPE_DIRECTORY);
       g_file_info_set_content_type (info, "inode/directory");
       g_file_info_set_name (info, "/");
-      g_vfs_backend_set_icon_name (backend, op_backend->icon_name);
+      if (op_backend->icon_name) {
+          GIcon *icon;
+
+          g_vfs_backend_set_icon_name (backend, op_backend->icon_name);
+          icon = g_themed_icon_new (op_backend->icon_name);
+          g_file_info_set_icon (info, icon);
+          g_object_unref (icon);
+      }
       display = g_strdup_printf (_("%s on %s"), "/", op_backend->display_name);
       g_file_info_set_display_name (info, display);
       g_free (display);
@@ -430,11 +706,18 @@ _query_file_info_helper (GVfsBackend *backend,
   g_list_foreach (elements, (GFunc)g_object_unref, NULL);
   g_list_free (elements);
 
-  g_print ("- _query_file_info_helper\n");
+  g_debug ("- _query_file_info_helper\n");
 
   return found;
 }
 
+static void
+_invalidate_cache_helper (GVfsBackendObexftp *op_backend)
+{
+  g_free (op_backend->directory);
+  op_backend->directory = NULL;
+}
+
 static void
 error_occurred_cb (DBusGProxy *proxy, const gchar *error_name, const gchar *error_message, gpointer user_data)
 {
@@ -539,15 +822,14 @@ do_mount (GVfsBackend *backend,
   const char *device;
   GError *error = NULL;
   const gchar *path = NULL;
-  char *server;
   GMountSpec *obexftp_mount_spec;
   guint count;
 
-  g_print ("+ do_mount\n");
+  g_debug ("+ do_mount\n");
 
   device = g_mount_spec_get (mount_spec, "host");
 
-  if (device == NULL || strlen (device) != BDADDR_LEN + 2)
+  if (device == NULL || (strlen (device) != BDADDR_LEN + 2 && !g_str_has_prefix(device, "[usb:")) )
     {
       g_vfs_job_failed (G_VFS_JOB (job),
                         G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
@@ -555,51 +837,91 @@ do_mount (GVfsBackend *backend,
       return;
     }
 
-  /* Strip the brackets */
-  op_backend->bdaddr = g_strndup (device + 1, 17);
-  if (bachk (op_backend->bdaddr) < 0)
+  op_backend->bdaddr = NULL;
+  op_backend->usbintfnum = -1;
+
+  if (!g_str_has_prefix(device, "[usb:"))
     {
-      g_free (op_backend->bdaddr);
-      g_vfs_job_failed (G_VFS_JOB (job),
-                        G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
-                        _("Invalid mount spec"));
-      return;
+      /* Strip the brackets */
+      op_backend->bdaddr = g_strndup (device + 1, BDADDR_LEN);
+      if (bachk (op_backend->bdaddr) < 0)
+        {
+          g_free (op_backend->bdaddr);
+          g_vfs_job_failed (G_VFS_JOB (job),
+                            G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
+                            _("Invalid mount spec"));
+          return;
+        }
+    }
+  else
+    {
+      op_backend->usbintfnum = _get_usb_intfnum_and_properties (op_backend->manager_proxy, device, &op_backend->display_name, &op_backend->icon_name);
+      if (op_backend->usbintfnum < 0)
+        {
+          if (op_backend->usbintfnum == -2)
+            {
+              g_vfs_job_failed (G_VFS_JOB (job),
+                                G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
+                                _("USB support missing. Please contact your software vendor"));
+            }
+          else
+            {
+              g_vfs_job_failed (G_VFS_JOB (job),
+                                G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
+                                _("Invalid mount spec"));
+            }
+          return;
+        }
     }
 
   /* FIXME, Have a way for the mount to be cancelled, see:
    * Use CancelSessionConnect */
   op_backend->status = ASYNC_PENDING;
 
-  if (dbus_g_proxy_call (op_backend->manager_proxy, "CreateBluetoothSession", &error,
-                         G_TYPE_STRING, op_backend->bdaddr, G_TYPE_STRING, "00:00:00:00:00:00", G_TYPE_STRING, "ftp", G_TYPE_INVALID,
-                         DBUS_TYPE_G_OBJECT_PATH, &path, G_TYPE_INVALID) == FALSE)
+  if (op_backend->bdaddr)
     {
-      g_free (op_backend->bdaddr);
-      g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
-      g_error_free (error);
-      return;
+      if (dbus_g_proxy_call (op_backend->manager_proxy, "CreateBluetoothSession", &error,
+                             G_TYPE_STRING, op_backend->bdaddr, G_TYPE_STRING, "00:00:00:00:00:00", G_TYPE_STRING, "ftp", G_TYPE_INVALID,
+                             DBUS_TYPE_G_OBJECT_PATH, &path, G_TYPE_INVALID) == FALSE)
+        {
+          g_free (op_backend->bdaddr);
+          g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
+          g_error_free (error);
+          return;
+        }
+      op_backend->display_name = _get_bluetooth_device_properties (op_backend->bdaddr, &op_backend->icon_name);
+      if (!op_backend->display_name)
+        op_backend->display_name = g_strdelimit (g_strdup (op_backend->bdaddr), ":", '-');
+      if (!op_backend->icon_name)
+        op_backend->icon_name = g_strdup ("bluetooth");
+      g_debug ("  do_mount: %s (%s) mounted\n", op_backend->display_name, op_backend->bdaddr);
+    }
+  else
+    {
+      if (dbus_g_proxy_call (op_backend->manager_proxy, "CreateUsbSession", &error,
+                             G_TYPE_UINT, op_backend->usbintfnum, G_TYPE_STRING, "ftp", G_TYPE_INVALID,
+                             DBUS_TYPE_G_OBJECT_PATH, &path, G_TYPE_INVALID) == FALSE)
+        {
+          g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
+          g_error_free (error);
+          return;
+        }
+      g_debug ("  do_mount: usb interface %d mounted\n", op_backend->usbintfnum);
     }
 
   g_vfs_job_set_backend_data (G_VFS_JOB (job), backend, NULL);
-  g_print ("  do_mount: %s mounted\n", op_backend->bdaddr);
 
   op_backend->session_proxy = dbus_g_proxy_new_for_name (op_backend->connection,
                                                          "org.openobex",
                                                          path,
                                                          "org.openobex.Session");
 
-  op_backend->display_name = _get_device_properties (op_backend->bdaddr, &op_backend->icon_name);
-  if (!op_backend->display_name)
-        op_backend->display_name = g_strdup (op_backend->bdaddr);
-
   g_vfs_backend_set_display_name (G_VFS_BACKEND  (op_backend),
                                   op_backend->display_name);
   g_vfs_backend_set_icon_name (G_VFS_BACKEND (op_backend), op_backend->icon_name);
 
   obexftp_mount_spec = g_mount_spec_new ("obex");
-  server = g_strdup_printf ("[%s]", op_backend->bdaddr);
-  g_mount_spec_set (obexftp_mount_spec, "host", server);
-  g_free (server);
+  g_mount_spec_set (obexftp_mount_spec, "host", device);
   g_vfs_backend_set_mount_spec (G_VFS_BACKEND (op_backend), obexftp_mount_spec);
   g_mount_spec_unref (obexftp_mount_spec);
 
@@ -625,6 +947,10 @@ do_mount (GVfsBackend *backend,
 
   dbus_g_proxy_add_signal(op_backend->session_proxy, "TransferStarted",
                           G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_INVALID);
+  dbus_g_proxy_add_signal(op_backend->session_proxy, "TransferCompleted",
+                          G_TYPE_INVALID);
+  dbus_g_proxy_add_signal(op_backend->session_proxy, "TransferProgress",
+                          G_TYPE_UINT64, G_TYPE_INVALID);
 
   /* Now wait until the device is connected */
   count = 0;
@@ -664,7 +990,7 @@ do_mount (GVfsBackend *backend,
 
   g_vfs_job_succeeded (G_VFS_JOB (job));
 
-  g_print ("- do_mount\n");
+  g_debug ("- do_mount\n");
 }
 
 static void
@@ -694,7 +1020,7 @@ do_open_for_read (GVfsBackend *backend,
   goffset size;
   int fd, success;
 
-  g_print ("+ do_open_for_read, filename: %s\n", filename);
+  g_debug ("+ do_open_for_read, filename: %s\n", filename);
 
   g_mutex_lock (op_backend->mutex);
   op_backend->doing_io = TRUE;
@@ -822,7 +1148,7 @@ do_open_for_read (GVfsBackend *backend,
   handle->size = size;
   g_vfs_job_open_for_read_set_handle (job, handle);
 
-  g_print ("- do_open_for_read, filename: %s\n", filename);
+  g_debug ("- do_open_for_read, filename: %s\n", filename);
 
   g_vfs_job_open_for_read_set_can_seek (G_VFS_JOB_OPEN_FOR_READ (job), FALSE);
   g_vfs_job_succeeded (G_VFS_JOB (job));
@@ -907,7 +1233,7 @@ do_close_read (GVfsBackend *backend,
   ObexFTPOpenHandle *backend_handle = (ObexFTPOpenHandle *) handle;
   int busy;
 
-  g_print ("+ do_close_read\n");
+  g_debug ("+ do_close_read\n");
 
   busy = is_busy (op_backend->session_proxy, G_VFS_JOB (job));
   if (busy < 0) {
@@ -937,7 +1263,7 @@ do_close_read (GVfsBackend *backend,
 
   g_vfs_job_succeeded (G_VFS_JOB (job));
 
-  g_print ("- do_close_read\n");
+  g_debug ("- do_close_read\n");
 }
 
 static void
@@ -951,7 +1277,7 @@ do_query_info (GVfsBackend *backend,
   GVfsBackendObexftp *op_backend = G_VFS_BACKEND_OBEXFTP (backend);
   GError *error = NULL;
 
-  g_print ("+ do_query_info, filename: %s\n", filename);
+  g_debug ("+ do_query_info, filename: %s\n", filename);
 
   g_mutex_lock (op_backend->mutex);
 
@@ -967,7 +1293,7 @@ do_query_info (GVfsBackend *backend,
 
   g_vfs_job_succeeded (G_VFS_JOB (job));
 
-  g_print ("- do_query_info\n");
+  g_debug ("- do_query_info\n");
 
   return;
 }
@@ -988,7 +1314,7 @@ do_query_fs_info (GVfsBackend *backend,
   GList *l;
   gboolean has_free_memory;
 
-  g_print ("+ do_query_fs_info, filename: %s\n", filename);
+  g_debug ("+ do_query_fs_info, filename: %s\n", filename);
 
   g_mutex_lock (op_backend->mutex);
 
@@ -1108,7 +1434,7 @@ set_info_from_memory:
 
   g_mutex_unlock (op_backend->mutex);
 
-  g_print ("- do_query_fs_info\n");
+  g_debug ("- do_query_fs_info\n");
 }
 
 static void
@@ -1123,7 +1449,7 @@ do_enumerate (GVfsBackend *backend,
   char *files;
   GList *elements = NULL;
 
-  g_print ("+ do_enumerate, filename: %s\n", filename);
+  g_debug ("+ do_enumerate, filename: %s\n", filename);
 
   g_mutex_lock (op_backend->mutex);
 
@@ -1176,7 +1502,327 @@ do_enumerate (GVfsBackend *backend,
 
   g_mutex_unlock (op_backend->mutex);
 
-  g_print ("- do_enumerate\n");
+  g_debug ("- do_enumerate\n");
+}
+
+typedef struct {
+  GVfsBackendObexftp *op_backend;
+  GFileProgressCallback progress_callback;
+  gpointer progress_callback_data;
+  goffset total_bytes;
+} PushData;
+
+static void
+push_transfer_started_cb (DBusGProxy *proxy,
+                          const gchar *filename,
+                          const gchar *local_path,
+                          guint64 total_bytes,
+                          gpointer user_data)
+{
+  PushData *job_data = (PushData *) user_data;
+  GVfsBackendObexftp *op_backend = job_data->op_backend;
+
+  g_message ("transfer of %s to %s started", filename, local_path);
+
+  g_mutex_lock (op_backend->mutex);
+
+  op_backend->status = ASYNC_RUNNING;
+  job_data->total_bytes = (goffset) total_bytes;
+  job_data->progress_callback (0, job_data->total_bytes,
+                               job_data->progress_callback_data);
+
+  g_cond_signal (op_backend->cond);
+  g_mutex_unlock (op_backend->mutex);
+}
+
+static void
+push_transfer_completed_cb (DBusGProxy *proxy,
+                            gpointer user_data)
+{
+  PushData *job_data = (PushData *) user_data;
+  GVfsBackendObexftp *op_backend = job_data->op_backend;
+
+  g_message ("transfer completed");
+
+  g_mutex_lock (op_backend->mutex);
+
+  op_backend->status = ASYNC_SUCCESS;
+
+  g_cond_signal (op_backend->cond);
+  g_mutex_unlock (op_backend->mutex);
+}
+
+static void
+push_transfer_progress_cb (DBusGProxy *proxy,
+                           guint64 bytes_transferred,
+                           gpointer user_data)
+{
+  PushData *job_data = (PushData *) user_data;
+
+  g_message ("transfer progress");
+
+  job_data->progress_callback ((goffset) bytes_transferred,
+                               job_data->total_bytes,
+                               job_data->progress_callback_data);
+}
+
+static void
+push_data_free (PushData *job_data)
+{
+  g_object_unref (job_data->op_backend);
+  g_slice_free (PushData, job_data);
+}
+
+static gboolean
+_push_single_file_helper (GVfsBackendObexftp *op_backend,
+                          GVfsJobPush *job,
+                          const char *local_path,
+                          const char *destination,
+                          GError **error,
+                          PushData *job_data)
+{
+  char *dirname;
+  int success;
+
+  dirname = g_path_get_dirname (destination);
+
+  if (_change_directory (op_backend, dirname, error) == FALSE)
+    {
+      g_free (dirname);
+      return FALSE;
+    }
+
+  g_free (dirname);
+
+  if (g_vfs_job_is_cancelled (G_VFS_JOB (job)))
+    {
+      g_set_error_literal (error, G_IO_ERROR,
+                           G_IO_ERROR_CANCELLED,
+                           _("Operation was cancelled"));
+      return FALSE;
+    }
+
+  op_backend->status = ASYNC_PENDING;
+
+  /* connect to the transfer signals */
+  dbus_g_proxy_connect_signal (op_backend->session_proxy, "TransferStarted",
+                               G_CALLBACK (push_transfer_started_cb), job_data,
+                               NULL);
+  dbus_g_proxy_connect_signal (op_backend->session_proxy, "TransferCompleted",
+                               G_CALLBACK (push_transfer_completed_cb), job_data,
+                               NULL);
+  dbus_g_proxy_connect_signal (op_backend->session_proxy, "TransferProgress",
+                               G_CALLBACK (push_transfer_progress_cb), job_data,
+                               NULL);
+
+  if (dbus_g_proxy_call (op_backend->session_proxy, "SendFile", error,
+                         G_TYPE_STRING, local_path, G_TYPE_INVALID,
+                         G_TYPE_INVALID) == FALSE)
+    {
+      dbus_g_proxy_disconnect_signal (op_backend->session_proxy, "TransferStarted", 
+                                      G_CALLBACK (push_transfer_started_cb), job_data);
+      dbus_g_proxy_disconnect_signal (op_backend->session_proxy, "TransferCompleted", 
+                                      G_CALLBACK (push_transfer_completed_cb), job_data);
+      dbus_g_proxy_disconnect_signal (op_backend->session_proxy, "TransferProgress", 
+                                      G_CALLBACK (push_transfer_progress_cb), job_data);
+      return FALSE;
+    }
+
+  /* wait for the TransferStarted or ErrorOccurred signal */
+  while (op_backend->status == ASYNC_PENDING)
+    g_cond_wait (op_backend->cond, op_backend->mutex);
+
+  dbus_g_proxy_disconnect_signal (op_backend->session_proxy, "TransferStarted",
+                                  G_CALLBACK (push_transfer_started_cb), job_data);
+  success = op_backend->status;
+
+  /* we either got the operation running or an error */
+  if (success == ASYNC_ERROR)
+    {
+      dbus_g_proxy_disconnect_signal (op_backend->session_proxy, "TransferCompleted", 
+                                      G_CALLBACK (push_transfer_completed_cb), job_data);
+      dbus_g_proxy_disconnect_signal (op_backend->session_proxy, "TransferProgress", 
+                                      G_CALLBACK (push_transfer_progress_cb), job_data);
+
+      *error = g_error_copy (op_backend->error);
+      g_error_free (op_backend->error);
+      op_backend->error = NULL;
+      return FALSE;
+    }
+
+  while (op_backend->status == ASYNC_RUNNING)
+    g_cond_wait (op_backend->cond, op_backend->mutex);
+
+  dbus_g_proxy_disconnect_signal (op_backend->session_proxy, "TransferCompleted", 
+                                  G_CALLBACK (push_transfer_completed_cb), job_data);
+  dbus_g_proxy_disconnect_signal (op_backend->session_proxy, "TransferProgress", 
+                                  G_CALLBACK (push_transfer_progress_cb), job_data);
+  success = op_backend->status;
+
+  /* same as before, either we have success or an error */
+  if (success == ASYNC_ERROR)
+    {
+      *error = g_error_copy (op_backend->error);
+      g_error_free (op_backend->error);
+      op_backend->error = NULL;
+      return FALSE;
+    }
+
+  return TRUE;
+}
+
+static void
+do_push (GVfsBackend *backend,
+         GVfsJobPush *job,
+         const char *destination,
+         const char *local_path,
+         GFileCopyFlags flags,
+         gboolean remove_source,
+         GFileProgressCallback progress_callback,
+         gpointer progress_callback_data)
+{
+  GVfsBackendObexftp *op_backend = G_VFS_BACKEND_OBEXFTP (backend);
+  GError *error = NULL;
+  gboolean is_dir, overwrite, res;
+  GFileType target_type;
+  PushData *job_data;
+  GFileInfo *info;
+
+  g_debug ("+ do_push, destination: %s, local_path: %s\n", destination, local_path);
+
+  g_mutex_lock (op_backend->mutex);
+  op_backend->doing_io = TRUE;
+
+  overwrite = (flags & G_FILE_COPY_OVERWRITE);
+  is_dir = g_file_test (local_path, G_FILE_TEST_IS_DIR);
+  info = g_file_info_new ();
+  target_type = 0;
+
+  if (g_vfs_job_is_cancelled (G_VFS_JOB (job)))
+    {
+      op_backend->doing_io = FALSE;
+      g_mutex_unlock (op_backend->mutex);
+
+      g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
+                        G_IO_ERROR_CANCELLED,
+                        _("Operation was cancelled"));
+      return;
+    }
+
+  res = _query_file_info_helper (backend, destination, info, &error);
+
+  if (error != NULL)
+    {
+      if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
+        {
+          op_backend->doing_io = FALSE;
+          g_mutex_unlock (op_backend->mutex);
+
+          g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
+          g_error_free (error);
+
+          return;
+        }
+      else
+        {
+          g_clear_error (&error);
+          res = TRUE;
+        }
+    }
+
+  if (res)
+    target_type = g_file_info_get_file_type (info);
+
+  g_object_unref (info);
+
+  /* error handling */
+  if (is_dir)
+    {
+      op_backend->doing_io = FALSE;
+      g_mutex_unlock (op_backend->mutex);
+
+      if (target_type != 0)
+        {
+          if (overwrite)
+            {
+              if (target_type == G_FILE_TYPE_DIRECTORY)
+                {
+                  g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
+                                    G_IO_ERROR_WOULD_MERGE,
+                                   _("Can't copy directory over directory"));
+                  return;
+                }
+            }
+          else
+            {
+              g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
+                                G_IO_ERROR_EXISTS,
+                                _("Target file exists"));
+              return;
+            }
+        }
+
+      g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
+                        G_IO_ERROR_WOULD_RECURSE,
+                        _("Can't recursively copy directory"));
+      return;
+    }
+  else
+    {
+      if (target_type != 0)
+        {
+          op_backend->doing_io = FALSE;
+          g_mutex_unlock (op_backend->mutex);
+
+          if (overwrite)
+            {
+              if (target_type == G_FILE_TYPE_DIRECTORY)
+                {
+                  g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
+                                    G_IO_ERROR_IS_DIRECTORY,
+                                    _("Can't copy file over directory"));
+                  return;
+                }
+            }
+          else
+            {
+              g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
+                                G_IO_ERROR_EXISTS,
+                                _("Target file exists"));
+              return;
+            }
+        }
+    }
+
+  job_data = g_slice_new0 (PushData);
+  job_data->op_backend = g_object_ref (op_backend);
+  job_data->progress_callback = progress_callback;
+  job_data->progress_callback_data = progress_callback_data;
+
+  /* start the actual transfer operation */
+  if (_push_single_file_helper (op_backend, job, local_path, destination,
+                                &error, job_data) == FALSE)
+    {
+      op_backend->doing_io = FALSE;
+      g_mutex_unlock (op_backend->mutex);
+      push_data_free (job_data);
+
+      g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
+    }
+
+  push_data_free (job_data);
+
+  /* we called _query_file_info_helper (), so we need to invalidate the
+   * cache, as a query_info () will be called on us after we return.
+   */
+  _invalidate_cache_helper (op_backend);
+
+  g_vfs_job_succeeded (G_VFS_JOB (job));
+
+  op_backend->doing_io = FALSE;
+  g_mutex_unlock (op_backend->mutex);
+
+  g_debug ("- do_push\n");
 }
 
 static void
@@ -1189,7 +1835,7 @@ do_delete (GVfsBackend *backend,
   GError *error = NULL;
   GFileInfo *info;
 
-  g_print ("+ do_delete, filename: %s\n", filename);
+  g_debug ("+ do_delete, filename: %s\n", filename);
 
   g_mutex_lock (op_backend->mutex);
 
@@ -1337,7 +1983,7 @@ do_delete (GVfsBackend *backend,
 
   g_mutex_unlock (op_backend->mutex);
 
-  g_print ("- do_delete\n");
+  g_debug ("- do_delete\n");
 }
 
 static void
@@ -1350,7 +1996,7 @@ do_make_directory (GVfsBackend *backend,
   GError *error = NULL;
   GFileInfo *info;
 
-  g_print ("+ do_make_directory, filename: %s\n", filename);
+  g_debug ("+ do_make_directory, filename: %s\n", filename);
 
   g_mutex_lock (op_backend->mutex);
 
@@ -1372,6 +2018,9 @@ do_make_directory (GVfsBackend *backend,
       return;
     }
 
+  if (error)
+    g_clear_error (&error);
+
   if (g_vfs_job_is_cancelled (G_VFS_JOB (job)))
     {
       g_mutex_unlock (op_backend->mutex);
@@ -1406,17 +2055,23 @@ do_make_directory (GVfsBackend *backend,
                          G_TYPE_INVALID) == FALSE)
     {
       g_mutex_unlock (op_backend->mutex);
+      g_free (basename);
       g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
       g_error_free (error);
       return;
     }
   g_free (basename);
 
+  /* reset the current directory so that we won't cache when querying
+   * info after this has succeeded.
+   */
+  _invalidate_cache_helper (op_backend);
+
   g_vfs_job_succeeded (G_VFS_JOB (job));
 
   g_mutex_unlock (op_backend->mutex);
 
-  g_print ("+ do_make_directory\n");
+  g_debug ("- do_make_directory\n");
 }
 
 static void
@@ -1436,6 +2091,7 @@ g_vfs_backend_obexftp_class_init (GVfsBackendObexftpClass *klass)
   backend_class->enumerate = do_enumerate;
   backend_class->delete = do_delete;
   backend_class->make_directory = do_make_directory;
+  backend_class->push = do_push;
 
   /* ErrorOccurred */
   dbus_g_object_register_marshaller (obexftp_marshal_VOID__STRING_STRING,
@@ -1444,6 +2100,9 @@ g_vfs_backend_obexftp_class_init (GVfsBackendObexftpClass *klass)
   /* TransferStarted */
   dbus_g_object_register_marshaller(obexftp_marshal_VOID__STRING_STRING_UINT64,
                                     G_TYPE_NONE, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_INVALID);
+  /* TransferProgress */
+  dbus_g_object_register_marshaller(obexftp_marshal_VOID__UINT64,
+                                    G_TYPE_NONE, G_TYPE_UINT64, G_TYPE_INVALID);
   /* SessionConnected */
   dbus_g_object_register_marshaller(obexftp_marshal_VOID__STRING,
                                     G_TYPE_NONE, DBUS_TYPE_G_OBJECT_PATH, G_TYPE_INVALID);
